import pkg_resources

from .question_answering import BertQuestionAnsweringModel

__version__ = pkg_resources.resource_string("adaptnlp", "VERSION.txt").decode("UTF-8").strip()

__all__ = ["__version__", "BertQuestionAnsweringModel",]



from typing import Tuple, List
from collections import OrderedDict

from adaptnlp.transformers import BertQuestionAnsweringModel


class EasyQuestionAnswering:
    """Question Answering Module
    """

    def __init__(self):
        self.bert_qa = None

    # Dynamic Loaders
    def _load_bert_qa(self) -> None:
        self.bert_qa = BertQuestionAnsweringModel()

    # QA Models
    def predict_bert_qa(self, query: str, context: str, n_best_size: int = 20) -> Tuple[str, List[OrderedDict]]:
        ''' Predicts top_n answer spans of query in regards to context


        :param query: The question
        :param context: The context of which the question is asking
        :param top_n: The top n answers returned

        :return: Either a list of string answers or a dict of the results
        '''

        self._load_bert_qa() if not self.bert_qa else None
        return self.bert_qa.predict(query=query, context=context, n_best_size=n_best_size)



import pkg_resources
from pathlib import Path

from .embeddings import EasyWordEmbeddings, EasyStackedEmbeddings, EasyDocumentEmbeddings
from .token_classification import EasyTokenTagger
from .sequence_classification import EasySequenceClassifier
from .question_answering import EasyQuestionAnswering
from .training import SequenceClassifierTrainer
from .transformers.finetuning import LMFineTuner

# global variable like flair's: cache_root
cache_root = Path(Path.home(), ".adaptnlp")

__version__ = pkg_resources.resource_string("adaptnlp", "VERSION.txt").decode("UTF-8").strip()

__all__ = ["__version__", "EasyWordEmbeddings", "EasyStackedEmbeddings", "EasyDocumentEmbeddings",
           "EasySequenceClassifier", "EasyTokenTagger", "EasyTextClassifierTrainer", "EasyQuestionAnswering",
           "SequenceClassifierTrainer", "LMFineTuner",]



from typing import Union, Dict
from pathlib import Path
import json
import csv

import numpy as np

from adaptnlp import EasyDocumentEmbeddings

from flair.datasets import CSVClassificationCorpus
from flair.data import Corpus
from flair.embeddings import DocumentRNNEmbeddings
from flair.models import TextClassifier
from flair.trainers import ModelTrainer
from flair.visual.training_curves import Plotter


class SequenceClassifierTrainer:

    def __init__(
            self,
            corpus: Union[Corpus, Path, str],
            encoder: Union[EasyDocumentEmbeddings, Path, str],
            column_name_map: None,
            corpus_in_memory: bool = True,
            predictive_head: str = "flair",
            **kwargs,
    ):
        """Sequence Classifier Trainer
    
        :param corpus: A flair corpus data model or `Path`/string to a directory with train.csv/test.csv/dev.csv
        :param encoder: A `EasyDocumentEmbeddings` object if training with a flair prediction head or `Path`/string if training with Transformer's prediction models
        :param column_name_map: Required if corpus is not a `Corpus` object, it's a dictionary specifying the indices of the text and label columns of the csv i.e. {1:"text",2:"label"}
        :param corpus_in_memory: Boolean for whether to store corpus embeddings in memory
        :param predictive_head: For now either "flair" or "transformers" for the prediction head
        """
        if isinstance(corpus, Corpus):
            self.corpus = corpus
        else:
            if isinstance(corpus, str):
                corpus = Path(corpus)
            if not column_name_map:
                raise ValueError(
                    "If not instantiating with `Corpus` object, must pass in `column_name_map` argument to specify text/label indices")
            self.corpus = CSVClassificationCorpus(corpus, column_name_map, skip_header=True, delimiter=",",
                                                  in_memory=corpus_in_memory)

        # Verify predictive head is within available heads
        self.available_predictive_head = ["flair", "transformers"]
        if predictive_head not in self.available_predictive_head:
            raise ValueError(f"predictive_head param must be one of the following: {self.available_predictive_head}")
        self.predictive_head = predictive_head

        # Verify correct corresponding encoder is used with predictive head (This can be structured with better design in the future)
        if isinstance(encoder, EasyDocumentEmbeddings):
            if predictive_head == "transformers":
                raise ValueError("If using `transformers` predictive head, pass in the path to the transformer's model")
            else:
                self.encoder = encoder
        else:
            if isinstance(encoder, str):
                encoder = Path(encoder)
            self.encoder = encoder

        # Create the label dictionary on init (store to keep from constantly generating label_dict) should we use dev/test set instead assuming all labels are provided?
        self.label_dict = self.corpus.make_label_dictionary()

        # Save trainer kwargs dict for reinitializations
        self.trainer_kwargs = kwargs

        # Load trainer with initial setup
        self._initial_setup(self.label_dict, **kwargs)

    def _initial_setup(self, label_dict: Dict, **kwargs):
        if self.predictive_head == "flair":

            # Get Document embeddings from `embeddings`
            document_embeddings: DocumentRNNEmbeddings = self.encoder.rnn_embeddings

            # Create the text classifier
            classifier = TextClassifier(
                document_embeddings,
                label_dictionary=label_dict,
                **kwargs,
            )

            # Initialize the text classifier trainer
            self.trainer = ModelTrainer(classifier, self.corpus)

        # TODO: In internal transformers package, create ****ForSequenceClassification adaptations
        elif self.predictive_head == "transformers":
            with open(self.encoder / "config.json") as config_f:
                configs = json.load(config_f)
                model_name = configs["architectures"][-1]
            if model_name == "BertForMaskedLM":
                pass

    def train(
            self,
            output_dir: Union[Path, str],
            learning_rate: float = 0.07,
            mini_batch_size: int = 32,
            anneal_factor: float = 0.5,
            patience: int = 5,
            max_epochs: int = 150,
            plot_weights: bool = False,
            **kwargs,
    ):
        """

        # TODO params

        :param output_dir:
        :param learning_rate:
        :param mini_batch_size:
        :param anneal_factor:
        :param patience:
        :param max_epochs:
        :param plot_weights:
        :param kwargs:
        """
        if isinstance(output_dir, str):
            output_dir = Path(output_dir)

        # Start the training
        self.trainer.train(
            output_dir,
            learning_rate=learning_rate,
            mini_batch_size=mini_batch_size,
            anneal_factor=anneal_factor,
            patience=patience,
            max_epochs=max_epochs,
            **kwargs,
        )

        # Plot weight traces
        if plot_weights:
            plotter = Plotter()
            plotter.plot_weights(output_dir / 'weights.txt')

    def find_learning_rate(
            self,
            output_dir: Union[Path, str],
            start_learning_rate: float = 1e-8,
            end_learning_rate: float = 10,
            iterations: int = 100,
            mini_batch_size: int = 32,
            stop_early: bool = True,
            smoothing_factor: float = 0.7,
            plot_learning_rate: bool = True,
            **kwargs,
    ) -> float:
        """
        # TODO params

        :param output_dir:
        :param start_learning_rate:
        :param end_learning_rate:
        :param iterations:
        :param mini_batch_size:
        :param stop_early:
        :param smoothing_factor:
        :param plot_learning_rate:
        :param kwargs:
        :return:
        """
        # 7. find learning rate
        learning_rate_tsv = self.trainer.find_learning_rate(base_path=output_dir,
                                                            file_name='learning_rate.tsv',
                                                            start_learning_rate=start_learning_rate,
                                                            end_learning_rate=end_learning_rate,
                                                            iterations=iterations,
                                                            mini_batch_size=mini_batch_size,
                                                            stop_early=stop_early,
                                                            smoothing_factor=smoothing_factor)

        # Reinitialize optimizer and parameters by reinitializing trainer
        self._initial_setup(self.label_dict, **self.trainer_kwargs)

        if plot_learning_rate:
            plotter = Plotter()
            plotter.plot_learning_rate(learning_rate_tsv)

        # Use the automated learning rate finder
        with open(learning_rate_tsv) as lr_f:
            lr_tsv = list(csv.reader(lr_f, delimiter="\t"))
        losses = np.array([float(row[-1]) for row in lr_tsv[1:]])
        lrs = np.array([float(row[-2]) for row in lr_tsv[1:]])
        lr_to_use = self._find_appropriate_lr(losses, lrs, **kwargs)
        print(f"Recommended Learning Rate {lr_to_use}")
        return lr_to_use

    def _find_appropriate_lr(
            self,
            losses: np.array,
            lrs: np.array,
            lr_diff: int = 15,
            loss_threshold: float = .2,
            adjust_value: float = 1,
    ) -> float:
        # Get loss values and their corresponding gradients, and get lr values
        assert (lr_diff < len(losses))
        loss_grad = np.gradient(losses)

        # Search for index in gradients where loss is lowest before the loss spike
        # Initialize right and left idx using the lr_diff as a spacing unit
        # Set the local min lr as -1 to signify if threshold is too low
        r_idx = -1
        l_idx = r_idx - lr_diff
        while (l_idx >= -len(losses)) and (abs(loss_grad[r_idx] - loss_grad[l_idx]) > loss_threshold):
            local_min_lr = lrs[l_idx]
            r_idx -= 1
            l_idx -= 1

        lr_to_use = local_min_lr * adjust_value

        return lr_to_use



from typing import List, Dict, Union
from collections import defaultdict

from flair.data import Sentence
from flair.models import SequenceTagger


class EasyTokenTagger:
    """ Token level classification models"""

    def __init__(self):
        self.token_taggers: Dict[SequenceTagger] = defaultdict(bool)

    def tag_text(
            self,
            text: Union[List[Sentence], Sentence, List[str], str],
            model_name_or_path: str = "ner-ontonotes",
            mini_batch_size: int = 32,
    ) -> List[Sentence]:
        """ Tags tokens with labels the token classification models have been trained on

        :param text: Text input, it can be a string or any of Flair's `Sentence` input formats
        :param model_name_or_path: The hosted model name key or model path
        :param mini_batch_size: The mini batch size for running inference
        :return: A list of Flair's `Sentence`s
        """
        # Load Sequence Tagger Model and Pytorch Module into tagger dict
        if not self.token_taggers[model_name_or_path]:
            self.token_taggers[model_name_or_path] = SequenceTagger.load(model_name_or_path)

        tagger = self.token_taggers[model_name_or_path]
        return tagger.predict(sentences=text, mini_batch_size=mini_batch_size, use_tokenizer=True)

    def tag_all(
            self,
            text: Union[List[Sentence], Sentence, List[str], str],
            mini_batch_size: int = 32,
    ) -> List[Sentence]:
        """ Tags tokens with all labels from all token classification models

        :param text: Text input, it can be a string or any of Flair's `Sentence` input formats
        :param mini_batch_size: The mini batch size for running inference
        :return: A list of Flair's `Sentence`s
        """
        if len(self.token_taggers) == 0:
            print("No token classification models loaded...")
            return Sentence()
        sentences = text
        for tagger_name in self.token_taggers.keys():
            sentences = self.tag_text(sentences, model_name_or_path=tagger_name, mini_batch_size=mini_batch_size)
        return sentences



#!/usr/bin/env python
from pathlib import Path
from setuptools import setup, find_packages

version_file = Path(__file__).parent.joinpath("adaptnlp", "VERSION.txt")
version = version_file.read_text(encoding="UTF-8").strip()

with open('requirements.txt') as reqs_file:
    install_requires = reqs_file.read().splitlines()

setup(
    name="adaptnlp",
    version=version,
    author="Andrew Chang",
    author_email="achang@novetta.com",
    packages=find_packages(),
    keywords=["NLP", "flair", "Natural Language Processing", "Machine Learning", "ML", "torch", "pytorch", "NER"],
    install_requires=install_requires,
    license="Apache",
    description="AdaptNLP: A Natural Language Processing Library and Framework",
    include_package_data=True,
    zip_safe=True
)



from typing import List, Dict, Union
from collections import defaultdict

from flair.data import Sentence
from flair.embeddings import (Embeddings,
                              WordEmbeddings,
                              StackedEmbeddings,
                              FlairEmbeddings,
                              BertEmbeddings,
                              DocumentPoolEmbeddings,
                              DocumentRNNEmbeddings,
                              OpenAIGPT2Embeddings,
                              XLNetEmbeddings,
                              XLMEmbeddings,
                              RoBERTaEmbeddings,
                              MuseCrosslingualEmbeddings)

FLAIR_PRETRAINED_MODEL_NAMES = {
    "multi-forward",
    "multi-backward",
    "multi-v0-forward",
    "multi-v0-backward",
    "multi-v0-forward-fast",
    "multi-v0-backward-fast",
    "en-forward",
    "en-backward",
    "en-forward-fast",
    "en-backward-fast",
    "news-forward",
    "news-backward",
    "news-forward-fast",
    "news-backward-fast",
    "mix-forward",
    "mix-backward",
    "ar-forward",
    "ar-backward",
    "bg-forward-fast",
    "bg-backward-fast",
    "bg-forward",
    "bg-backward",
    "cs-forward",
    "cs-backward",
    "cs-v0-forward",
    "cs-v0-backward",
    "da-forward",
    "da-backward",
    "de-forward",
    "de-backward",
    "de-historic-ha-forward",
    "de-historic-ha-backward",
    "de-historic-wz-forward",
    "de-historic-wz-backward",
    "es-forward",
    "es-backward",
    "es-forward-fast",
    "es-backward-fast",
    "eu-forward",
    "eu-backward",
    "eu-v1-forward",
    "eu-v1-backward",
    "eu-v0-forward",
    "eu-v0-backward",
    "fa-forward",
    "fa-backward",
    "fi-forward",
    "fi-backward",
    "fr-forward",
    "fr-backward",
    "he-forward",
    "he-backward",
    "hi-forward",
    "hi-backward",
    "hr-forward",
    "hr-backward",
    "id-forward",
    "id-backward",
    "it-forward",
    "it-backward",
    "ja-forward",
    "ja-backward",
    "nl-forward",
    "nl-backward",
    "nl-v0-forward",
    "nl-v0-backward",
    "no-forward",
    "no-backward",
    "pl-forward",
    "pl-backward",
    "pl-opus-forward",
    "pl-opus-backward",
    "pt-forward",
    "pt-backward",
    "pubmed-forward",
    "pubmed-backward",
    "sl-forward",
    "sl-backward",
    "sl-v0-forward",
    "sl-v0-backward",
    "sv-forward",
    "sv-backward",
    "sv-v0-forward",
    "sv-v0-backward",
    "ta-forward",
    "ta-backward",
}


class EasyWordEmbeddings:
    """ Word embeddings from the latest language models
    """

    def __init__(self):
        self.models: Dict[Embeddings] = defaultdict(bool)

    def embed_text(
            self,
            text: Union[List[Sentence], Sentence, List[str], str],
            model_name_or_path: str = "bert-base-cased",
    ) -> List[Sentence]:
        """ Produces embeddings for text

        :param text: Text input, it can be a string or any of Flair's `Sentence` input formats
        :param model_name_or_path: The hosted model name key or model path
        :return: A list of Flair's `Sentence`s
        """
        # Convert into sentences
        if isinstance(text, str):
            sentences = Sentence(text)
        elif isinstance(text, list) and all(isinstance(t, str) for t in text):
            sentences = [Sentence(t) for t in text]
        else:
            sentences = text

        # Load correct Embeddings module
        if not self.models[model_name_or_path]:
            if "bert" in model_name_or_path and "roberta" not in model_name_or_path:
                self.models[model_name_or_path] = BertEmbeddings(model_name_or_path)
            elif "roberta" in model_name_or_path:
                self.models[model_name_or_path] = RoBERTaEmbeddings(model_name_or_path)
            elif "gpt2" in model_name_or_path:
                self.models[model_name_or_path] = OpenAIGPT2Embeddings(model_name_or_path)
            elif "xlnet" in model_name_or_path:
                self.models[model_name_or_path] = XLNetEmbeddings(model_name_or_path)
            elif "xlm" in model_name_or_path:
                self.models[model_name_or_path] = XLMEmbeddings(model_name_or_path)
            elif "flair" in model_name_or_path or model_name_or_path in FLAIR_PRETRAINED_MODEL_NAMES:
                self.models[model_name_or_path] = FlairEmbeddings(model_name_or_path)
            else:
                print(f"Corresponding flair embedding module not found for {model_name_or_path}")
                return Sentence("")
        embedding = self.models[model_name_or_path]
        return embedding.embed(sentences)

    def embed_all(
            self,
            text: Union[List[Sentence], Sentence, List[str], str],
            *model_names_or_paths: str,
    ) -> List[Sentence]:
        """Embeds text with all embedding models loaded

        :param text: Text input, it can be a string or any of Flair's `Sentence` input formats
        :param model_names_or_paths: A variable input of model names or paths to embed
        :return: A list of Flair's `Sentence`s
        """
        # Convert into sentences
        if isinstance(text, str):
            sentences = Sentence(text)
        elif isinstance(text, list) and all(isinstance(t, str) for t in text):
            sentences = [Sentence(t) for t in text]
        else:
            sentences = text

        if model_names_or_paths:
            for embedding_name in model_names_or_paths:
                sentences = self.embed_text(sentences, model_name_or_path=embedding_name)
        else:
            for embedding_name in self.models.keys():
                sentences = self.embed_text(sentences, model_name_or_path=embedding_name)
        return sentences


class EasyStackedEmbeddings:
    """ Word Embeddings that have been concatenated and "stacked" as specified by flair
    """

    def __init__(self, *embeddings: str):
        '''

        :param embeddings: Strings specifying the embeddings you want to stack
        '''
        print("May need a couple moments to instantiate...")
        self.embedding_stack = []

        # Load correct Embeddings module
        for model_name_or_path in embeddings:
            if "bert" in model_name_or_path and "roberta" not in model_name_or_path:
                self.embedding_stack.append(BertEmbeddings(model_name_or_path))
            elif "roberta" in model_name_or_path:
                self.embedding_stack.append(RoBERTaEmbeddings(model_name_or_path))
            elif "gpt2" in model_name_or_path:
                self.embedding_stack.append(OpenAIGPT2Embeddings(model_name_or_path))
            elif "xlnet" in model_name_or_path:
                self.embedding_stack.append(XLNetEmbeddings(model_name_or_path))
            elif "xlm" in model_name_or_path:
                self.embedding_stack.append(XLMEmbeddings(model_name_or_path))
            elif "flair" in model_name_or_path or model_name_or_path in FLAIR_PRETRAINED_MODEL_NAMES:
                self.embedding_stack.append(FlairEmbeddings(model_name_or_path))
            else:
                print(f"Corresponding flair embedding module not found for {model_name_or_path}")

        assert (len(self.embedding_stack) != 0)
        self.stacked_embeddings = StackedEmbeddings(embeddings=self.embedding_stack)

    def embed_text(
            self,
            text: Union[List[Sentence], Sentence, List[str], str],
    ) -> List[Sentence]:
        """ Stacked embeddings

        :param text: Text input, it can be a string or any of Flair's `Sentence` input formats
        :return: A list of Flair's `Sentence`s
        """
        # Convert into sentences
        if isinstance(text, str):
            sentences = [Sentence(text)]
        elif isinstance(text, list) and all(isinstance(t, str) for t in text):
            sentences = [Sentence(t) for t in text]
        elif isinstance(text, Sentence):
            sentences = [text]
        else:
            sentences = text

        # Unlike flair embeddings modules, stacked embeddings do not return a list of sentences
        self.stacked_embeddings.embed(sentences)
        return sentences


class EasyDocumentEmbeddings():
    """ Document Embeddings generated by pool and rnn methods applied to the word embeddings of text
    """

    __allowed_methods = ["rnn", "pool"]
    __allowed_configs = ("pool_configs", "rnn_configs")

    def __init__(
            self,
            *embeddings: str,
            methods: List[str] = ["rnn", "pool"],
            configs: Dict = {"pool_configs": {"fine_tune_mode": "linear", "pooling": "mean", },
                             "rnn_configs": {"hidden_size": 512,
                                             "rnn_layers": 1,
                                             "reproject_words": True,
                                             "reproject_words_dimension": 256,
                                             "bidirectional": False,
                                             "dropout": 0.5,
                                             "word_dropout": 0.0,
                                             "locked_dropout": 0.0,
                                             "rnn_type": "GRU",
                                             "fine_tune": True, },
                             },
    ):
        '''

        :param embeddings: A variable number of strings referring to model names or paths
        :param methods: A list of strings to specify which document embeddings to use i.e. ["rnn", "pool"] (avoids unncessary loading of models if only using one)
        :param configs: A dictionary of configurations for flair's rnn and pool document embeddings
        '''
        print("May need a couple moments to instantiate...")
        self.embedding_stack = []

        # Check methods
        for m in methods:
            assert (m in self.__class__.__allowed_methods)

        # Set configs for pooling and rnn parameters
        for k, v in configs.items():
            assert (k in self.__class__.__allowed_configs)
            setattr(self, k, v)

        # Load correct Embeddings module
        for model_name_or_path in embeddings:
            if "bert" in model_name_or_path and "roberta" not in model_name_or_path:
                self.embedding_stack.append(BertEmbeddings(model_name_or_path))
            elif "roberta" in model_name_or_path:
                self.embedding_stack.append(RoBERTaEmbeddings(model_name_or_path))
            elif "gpt2" in model_name_or_path:
                self.embedding_stack.append(OpenAIGPT2Embeddings(model_name_or_path))
            elif "xlnet" in model_name_or_path:
                self.embedding_stack.append(XLNetEmbeddings(model_name_or_path))
            elif "xlm" in model_name_or_path:
                self.embedding_stack.append(XLMEmbeddings(model_name_or_path))
            elif "flair" in model_name_or_path or model_name_or_path in FLAIR_PRETRAINED_MODEL_NAMES:
                self.embedding_stack.append(FlairEmbeddings(model_name_or_path))
            else:
                print(f"Corresponding flair embedding module not found for {model_name_or_path}")

        assert (len(self.embedding_stack) != 0)
        if "pool" in methods:
            self.pool_embeddings = DocumentPoolEmbeddings(self.embedding_stack, **self.pool_configs)
            print("Pooled embedding loaded")
        if "rnn" in methods:
            self.rnn_embeddings = DocumentRNNEmbeddings(self.embedding_stack, **self.rnn_configs)
            print("RNN embeddings loaded")

    def embed_pool(
            self,
            text: Union[List[Sentence], Sentence, List[str], str],
    ) -> List[Sentence]:
        """ Stacked embeddings

        :param text: Text input, it can be a string or any of Flair's `Sentence` input formats
        :return: A list of Flair's `Sentence`s
        """
        if isinstance(text, str):
            sentences = [Sentence(text)]
        elif isinstance(text, list) and all(isinstance(t, str) for t in text):
            sentences = [Sentence(t) for t in text]
        elif isinstance(text, Sentence):
            sentences = [text]
        else:
            sentences = text
        self.pool_embeddings.embed(sentences)
        return sentences

    def embed_rnn(
            self,
            text: Union[List[Sentence], Sentence, List[str], str],
    ) -> List[Sentence]:
        """ Stacked embeddings

        :param text: Text input, it can be a string or any of Flair's `Sentence` input formats
        :return: A list of Flair's `Sentence`s
        """
        if isinstance(text, str):
            sentences = [Sentence(text)]
        elif isinstance(text, list) and all(isinstance(t, str) for t in text):
            sentences = [Sentence(t) for t in text]
        elif isinstance(text, Sentence):
            sentences = [text]
        else:
            sentences = text
        self.rnn_embeddings.embed(sentences)
        return sentences
